//
// This file is part of the Marble Virtual Globe.
//
// This program is free software licensed under the GNU LGPL. You can
// find a copy of this license in LICENSE.txt in the top directory of
// the source code.
//
// Copyright 2011 Florian Eßer <f.esser@rwth-aachen.de>
//

#include "ElevationDataGenerator.h"

// #include <QThread>
#include "MarbleDebug.h"
// #include "MarbleLocale.h"
#include "MarbleModel.h"
// #include "GeoPainter.h"
// #include "ViewportParams.h"
// #include "routing/RoutingManager.h"
#include "MarbleDirs.h"

namespace Marble
{

ElevationDataGenerator::ElevationDataGenerator()
    : m_eleData( QList<QPointF>() ),
      m_points( GeoDataLineString() ),
      m_maxElevation( 0 ),
      m_missingHgtFiles( QStringList() ),
      m_storagePolicy( 0 ),
      m_downloadManager( 0 ),
      m_unzipProcess( 0 ),
      m_currentZipFileName( QString() ),
      m_unzipQueue( QStringList() ),
      m_marbleWidget( 0 )
{
}

ElevationDataGenerator::~ElevationDataGenerator()
{
}

void ElevationDataGenerator::setMarbleWidget(MarbleWidget* widget)
{
    m_marbleWidget = widget;
}



void ElevationDataGenerator::calculateDistances()
{
    // TODO: Don't re-calculate the whole route if only a small part of it was changed
    m_eleData.clear();

    GeoDataLineString *path = new GeoDataLineString();
    for ( int i = 0; i < m_points.size(); i++ ) {
        path->append(m_points[i]);
        m_eleData.append( QPointF( path->length(EARTH_RADIUS), m_points[i].altitude() ) );
    }
}



void ElevationDataGenerator::calculateElevations()
{
    qreal ele;
    QString hgtFilename;
    QList<qint16> hgtData;

    for ( int i = 0; i < m_eleData.size(); i++ ) {
        if ( m_eleData[i].y() != 0 ) {
            // already has a value
            continue;
        }
        if ( hgtFilename != latLonToHgtFilename( m_points[i] ) ) {
            // we entered a new hgt file's area --> load new file
            hgtFilename = latLonToHgtFilename( m_points[i] );
            if ( !m_missingHgtFiles.contains( hgtFilename ) ) {

				QString srtm = latlongToSrtmArea( m_points[i] );
                hgtData = readHgtFile( hgtFilename,srtm );
            }
        }

        ele = lookupElevation( m_points[i], hgtData );
        if( ele == -32768) { // SRTM void
            ele = 0;
        }
        m_eleData[i].setY( ele );
    }

    QPointF point;
    m_maxElevation = -1;
    foreach ( point, m_eleData ) {
        if ( point.y() > m_maxElevation ) {
            m_maxElevation = point.y();
        }
    }
    emit dataUpdated();
}



QString ElevationDataGenerator::latLonToHgtFilename( const GeoDataCoordinates& coordinates )
{
    qreal lat = coordinates.latitude ( Marble::GeoDataCoordinates::Degree );
    qreal lon = coordinates.longitude( Marble::GeoDataCoordinates::Degree );

	if(lon>0 && lat>0)
		return QString("N%1E%2.hgt").
                arg( (int) lat, 2, 10, QChar( '0' ) ).
                arg( (int) lon, 3, 10, QChar( '0' ) );
	if(lon>0 && lat<0)
		return QString("S%1E%2.hgt").
                arg( (int) lat, 2, 10, QChar( '0' ) ).
                arg( (int) lon, 3, 10, QChar( '0' ) );
	if(lon<0 && lat>0)
		return QString("N%1W%2.hgt").
                arg( (int) lat, 2, 10, QChar( '0' ) ).
                arg( (int) lon, 3, 10, QChar( '0' ) );
	if(lon<0 && lat<0)
		return QString("S%1W%2.hgt").
                arg( (int) lat, 2, 10, QChar( '0' ) ).
                arg( (int) lon, 3, 10, QChar( '0' ) );
}



QList<qint16> ElevationDataGenerator::readHgtFile( const QString &filename, const QString &srtm )
{
    QString path = MarbleDirs::path( "srtm/" + filename);

    QList<qint16> contents;
    QFile hgtFile( path );
    if ( path.isEmpty() || !hgtFile.open( QIODevice::ReadOnly ) ) {
        if ( !m_missingHgtFiles.contains( filename ) ) {
            m_missingHgtFiles.append( filename );
            if ( !m_downloadManager ) {
                // initialize download manager
                m_storagePolicy = new FileStoragePolicy( MarbleDirs::localPath() + "/srtm/" );
                m_downloadManager = new HttpDownloadManager( m_storagePolicy, m_marbleWidget->model()->pluginManager() );
                connect( m_downloadManager, SIGNAL( downloadComplete( QString, QString ) ),
                                            SLOT ( addToUnzipQueue ( QString, QString ) ) );
            }
            // FIXME: Somehow detect the continent
            QString url( srtm + filename + ".zip" );
			//QString url("http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Eurasia/"+ filename + ".zip" );
            m_downloadManager->addJob( url, filename + ".zip", filename, DownloadBrowse );
        }
    } else {
        QDataStream input( &hgtFile );
        input.setByteOrder( QDataStream::BigEndian );
        qint16 n;
        while ( ! input.atEnd() ) {
            input >> n;
            contents.append( n );
        }
        hgtFile.close();
    }

    return contents;
}



void ElevationDataGenerator::addToUnzipQueue( QString filename, QString downloadId )
{
    Q_UNUSED( downloadId );
    m_unzipQueue.append( filename );
    mDebug() << "appending " << filename << " to unzip queue. "
             << "Queue is now: [" << m_unzipQueue.join(", ") << "]";
    unzipHgt();
}



void ElevationDataGenerator::unzipHgt()
{
    if ( m_unzipQueue.isEmpty() ) {
        return;
    }

    if ( m_unzipProcess ) {
        // another unzip process is still running
        return;
    } else if ( canExecute( "unzip.exe" ) ) {
        QString filename = m_unzipQueue.first();
        m_currentZipFileName = MarbleDirs::localPath() + "/srtm/" + filename;

        QFileInfo file( m_currentZipFileName );
        if ( file.exists() && file.isReadable() ) {
            m_unzipProcess = new QProcess;
            connect( m_unzipProcess, SIGNAL( finished( int ) ), SLOT ( unzipFinished( int ) ) );

            QStringList arguments = QStringList() << m_currentZipFileName;;//file.fileName();
			QString qsApp = MarbleDirs::localPath() + "/srtm/bin/unzip.exe"; 
            m_unzipProcess->setWorkingDirectory( file.dir().absolutePath() );
            m_unzipProcess->start(qsApp, arguments );
            mDebug() << "unzip process started";
        } else {
            mDebug() << "Unzip error: Could not open file " << m_currentZipFileName << "(" << filename << ")";
        }
    } else {
        mDebug() << "Cannot extract archive: unzip executable not found in PATH.";
    }
}



void ElevationDataGenerator::unzipFinished( int exitCode )
{
    mDebug() << "unzip process finished with exit code " << QString::number(exitCode);
    if ( exitCode ) {
        mDebug() << "Unzip error!";
        return;
    }
    m_missingHgtFiles.removeOne( QFileInfo( m_currentZipFileName ).completeBaseName() );
    if ( ! QFile::remove( m_currentZipFileName ) ) {
        mDebug() << "Error: Could not remove Zip file " << m_currentZipFileName;
    }
    m_currentZipFileName = QString();

    if ( m_unzipProcess ) {
        m_unzipProcess->close();
        m_unzipProcess->deleteLater();
        m_unzipProcess = 0;
    }
    if ( !m_unzipQueue.isEmpty() ) {
        QString nextFile = m_unzipQueue.first();
        m_unzipQueue.removeFirst();
        unzipHgt();
    }

    calculateElevations();
}



qreal ElevationDataGenerator::lookupElevation( const Marble::GeoDataCoordinates& coordinates,
                                                 const QList<qint16>& hgtData )
{
    qreal lat = coordinates.latitude ( Marble::GeoDataCoordinates::Degree );
    qreal lon = coordinates.longitude( Marble::GeoDataCoordinates::Degree );

    const int hgtGridSize = 1201;

    // TODO: Interpolation
    int x = ( lon - (int) lon ) * hgtGridSize;
    int y = hgtGridSize - ( lat - (int) lat ) * hgtGridSize - 1;
    return hgtData.value( y * hgtGridSize + x );
}



// copied from MonavConfigWidgetPrivate::canExecute
bool ElevationDataGenerator::canExecute( const QString &executable ) const
{
 /*   QString path = QProcessEnvironment::systemEnvironment().value( "PATH", "/usr/local/bin:/usr/bin:/bin" );
    foreach( const QString &dir, path.split( ":" ) ) {
        QFileInfo application( QDir( dir ), executable );
        if ( application.exists() ) {
            return true;
        }
    }*/

	QString qsDir = MarbleDirs::localPath() + "/srtm/bin/";
	QFileInfo application( QDir( qsDir ), executable );
    if ( application.exists() ) {
        return true;
    }

    return false;
}

QString	ElevationDataGenerator::latlongToSrtmArea( const GeoDataCoordinates &coordinates )
{
	int ret=0;//1=Africa,2=Australia,3=Eurasia,4=Islands,5=NA,6=SA

	QString srtm;

	qreal qLat = coordinates.latitude ( Marble::GeoDataCoordinates::Degree );
    qreal qLon = coordinates.longitude( Marble::GeoDataCoordinates::Degree );

	int lon = (int)qLon;
	int lat = (int)qLat;

	//eurasia
	if(lon >= -14 && lat <= 61 && lat >= 36 && lon <=180)
		ret=3;
	else if(lon >= 61 && lat <= 36 && lat >= -11 && lon <=180)
		ret=3;

	//africa
	else if(lon >= -26 && lat <= 35 && lat >= -35 && lon <=60)
		ret=1;

	//NA
	else if(lon >= -170 && lat <= 61 && lat >= 15 && lon <= -42)
		ret=5;

	//SA
	else if(lon >= -95 && lat <= 15 && lat >= -56 && lon <= -32)
		ret=6;

	//Australia
	else if(lon >= 113 && lat <= -10 && lat >= -45 && lon <= 180)
		ret=2;

	else ret=4;
	
	switch(ret)
	{
		case 1:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Africa/";
				break;
		case 2:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Australia/";
				break;
		case 3:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Eurasia/";
				break;
		case 4:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Islandes/";
				break;
		case 5:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/North_America/";
				break;
		case 6:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/South_America/";
				break;
		default:
				srtm="http://dds.cr.usgs.gov/srtm/version2_1/SRTM3/Eurasia/";
				break;
	}

	return srtm;

}

}

#include "./temp/moc_ElevationDataGenerator.cpp"